import sys
from argparse import ArgumentParser
import json
import pathlib
import pandas as pd
import joblib
import multiprocessing
import keras
import pathlib
import requests

import numpy as np

import allestm.parsers
import allestm.utils
import xgboost as xgb

from allestm.features.categorical import Sequence
from allestm.features.continuous import Pssm

from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LinearRegression, LogisticRegression

import logging

logging.basicConfig(level=logging.INFO)
log = logging.getLogger()


def parse_args(argv):
    parser = ArgumentParser(description='')
    parser.add_argument('a3m', help='MSA in a3m format.')
    parser.add_argument('out', help='Output file (JSON).')

    parser.add_argument('-m', '--model-dir', help='Model dir.')
    parser.add_argument('-c', '--config', help='Config file.')

    parser.add_argument('-d', '--decimals', type=int, default=3, help='Number of decimals the final result is rounded to.')
    parser.add_argument('-l', '--no-dl', action='store_true', help='Do NOT use deep learning models')
    parser.add_argument('-r', '--no-rf', action='store_true', help='Do NOT use rf models')
    parser.add_argument('-x', '--no-xgb', action='store_true', help='Do NOT use xgb models')
    return parser.parse_args(argv)


def prepare_data(seq, pssm, window=0):
    feature_dfs = []
    for feature, result in [(Sequence(), seq), (Pssm(), pssm)]:
        log.info(f'Calculating feature {type(feature)}.')

        df = pd.DataFrame(result, columns=[f"{type(feature).__qualname__}_{i}" for i in range(result.shape[1])]) 
        if window != 0:
            for column in df.columns:
                for win in range(-int((window - 1) / 2), int((window + 1) / 2)):
                    if win != 0:
                        new_column = f"{column}_w{win}"
                        df[new_column] = df[column].shift(win)

        feature_dfs.append(df.ffill().bfill())

    return feature_dfs


def download_model(s3_path, model_dir, model_filename):
    if not (model_dir / model_filename).exists():
        print(f'Model {model_filename} not present in {model_dir}, downloading from {s3_path}.')
        url = f'{s3_path}{model_filename}'
        r = requests.get(url)
        with open(model_dir / model_filename, 'wb') as fh:
            fh.write(r.content)


def main():
    args = parse_args(sys.argv[1:])

    if args.config is None:
        config_file = pathlib.Path(__file__).parent / 'config.json'
    else:
        config_file = pathlib.Path(args.config)

    # Parsing config.
    log.debug('Parsing config file.')
    with open(config_file, 'r') as config_fh:
        config = json.load(config_fh)

    if args.config is None:
        model_dir = pathlib.Path(__file__).parent / 'models/'
        model_dir.mkdir(parents=True, exist_ok=True)
    else:
        model_dir = args.model_dir

    # Parsing a3m file.
    log.debug('Parsing a3m.')
    fasta, a3m = allestm.parsers.parse_a3m(args.a3m)

    seq_transformer = allestm.features.categorical.Sequence()
    seq = seq_transformer.transform(fasta)

    pssm_transformer = allestm.features.continuous.Pssm()
    pssm = pssm_transformer.transform(a3m)

    results = {}

    # DL.
    if not args.no_dl:
        def run_dl(seq, pssm, model_file, queue):
            model = keras.models.load_model(model_file)
            predictions = model.predict([dl_seq_data, dl_pssm_data])
            queue.put(predictions)

        dl_seq_data_tmp, dl_pssm_data_tmp = (x.to_numpy() for x in prepare_data(seq, pssm, window=0))

        dl_seq_data = np.zeros((1, config['config']['max_length'], len(seq_transformer)))
        dl_pssm_data = np.zeros((1, config['config']['max_length'], len(seq_transformer)))

        dl_seq_data[0, :dl_seq_data_tmp.shape[0], :dl_seq_data_tmp.shape[1]] = dl_seq_data_tmp
        dl_pssm_data[0, :dl_pssm_data_tmp.shape[0], :dl_pssm_data_tmp.shape[1]] = dl_pssm_data_tmp

        for method in ['cnn', 'dcnn', 'lstm']:
            for target_group in config['methods'][method]:
                log.debug(f'Target group {target_group}.')

                for i, model_config in enumerate(config['methods'][method][target_group]):
                    print(f'Running model {i+1} of {len(config["methods"][method][target_group])} for method {method} and target {model_config["targets"]}')

                    log.debug(f'Model config: {model_config}.')
                    queue = multiprocessing.SimpleQueue()

                    model_file = model_dir / model_config["model_file"]
                    download_model(config['config']['s3_path'], model_dir, model_config['model_file'])
                    log.debug(f'Loading model file {model_file} for targets {model_config["targets"]}.')

                    # Necessary to free memory after prediction.
                    p = multiprocessing.Process(target=run_dl, args=(dl_seq_data, dl_pssm_data, model_file, queue))
                    p.start()
                    p.join()

                    predictions = queue.get()
                    for i, target_name in enumerate(model_config["targets"].split(',')):
                        target = allestm.utils.name_to_feature(f'allestm.features.{target_name}')

                        if len(predictions[i].shape) > 2:
                            preds = predictions[i][0, :dl_seq_data_tmp.shape[0], :]
                        else:
                            preds = predictions[i][:dl_seq_data_tmp.shape[0], :]

                        if predictions[i].shape[-1] == 1:
                            preds = preds.squeeze()

                        if hasattr(target, 'inverse_transform'):
                            preds = target.inverse_transform(preds)

                        results.setdefault(target_name, {}).setdefault(method, {}).setdefault(model_config['fold'], preds.tolist())

    # RF.
    if not args.no_rf:
        for target_group in config['methods']['rf']:
            log.debug(f'Target group {target_group}.')

            for i, model_config in enumerate(config['methods']['rf'][target_group]):
                print(f'Running model {i+1} of {len(config["methods"]["rf"][target_group])} for method rf and target {model_config["targets"]}')

                log.debug(f'Model config: {model_config}.')
                data = pd.concat(prepare_data(seq, pssm, window=model_config['window_size']), axis=1)

                target = allestm.utils.name_to_feature(f'allestm.features.{model_config["targets"]}')

                model_file = model_dir / model_config["model_file"]
                download_model(config['config']['s3_path'], model_dir, model_config['model_file'])
                log.debug(f'Loading model file {model_file} for target {target}.')
                model = joblib.load(model_file)

                if isinstance(model, RandomForestRegressor):
                    predictions = model.predict(data)
                else:
                    predictions = model.predict_proba(data)

                if isinstance(target, allestm.features.binary.Binary):
                    predictions = predictions[:, 1]

                if hasattr(target, 'inverse_transform'):
                    predictions = target.inverse_transform(predictions)

                results.setdefault(model_config['targets'], {}).setdefault('rf', {}).setdefault(model_config['fold'], predictions.tolist())

    # XGB.
    if not args.no_xgb:
        for target_group in config['methods']['xgb']:
            log.debug(f'Target group {target_group}.')

            for i, model_config in enumerate(config['methods']['xgb'][target_group]):
                print(f'Running model {i+1} of {len(config["methods"]["xgb"][target_group])} for method xgb and target {model_config["targets"]}')

                log.debug(f'Model config: {model_config}.')
                data = pd.concat(prepare_data(seq, pssm, window=model_config['window_size']), axis=1)

                target = allestm.utils.name_to_feature(f'allestm.features.{model_config["targets"]}')

                model_file = model_dir / model_config["model_file"]
                download_model(config['config']['s3_path'], model_dir, model_config['model_file'])
                log.debug(f'Loading model file {model_file} for target {target}.')

                model = xgb.Booster({'nthread': multiprocessing.cpu_count()})
                model.load_model(model_file)

                predictions = model.predict(xgb.DMatrix(data), ntree_limit=int(model.attributes()['best_iteration']) + 1)

                if hasattr(target, 'inverse_transform'):
                    predictions = target.inverse_transform(predictions)

                results.setdefault(model_config['targets'], {}).setdefault('xgb', {}).setdefault(model_config['fold'], predictions.tolist())

    #with open('tmp.json', 'r') as fh:
        #results = json.load(fh)

    # Blending
    if not args.no_dl and not args.no_rf and not args.no_xgb:
        for target_group in config['methods']['blending']:
            log.debug(f'Target group {target_group}.')

            for i, model_config in enumerate(config['methods']['blending'][target_group]):
                print(f'Running model {i+1} of {len(config["methods"]["blending"][target_group])} for method blending and target {model_config["targets"]}')

                log.debug(f'Model config: {model_config}.')
                data = np.column_stack([np.array(results[model_config['targets']][method][str(model_config['fold'])]) for method in ['cnn', 'dcnn', 'lstm', 'rf', 'xgb']])

                target = allestm.utils.name_to_feature(f'allestm.features.{model_config["targets"]}')

                model_file = model_dir / model_config["model_file"]
                download_model(config['config']['s3_path'], model_dir, model_config['model_file'])
                log.debug(f'Loading model file {model_file} for target {target}.')
                model = joblib.load(model_file)

                if isinstance(model, LinearRegression):
                    predictions = model.predict(data)
                else:
                    predictions = model.predict_proba(data)

                results.setdefault(model_config['targets'], {}).setdefault('blending', {}).setdefault(model_config['fold'], predictions.tolist())

        # Avg
        for target in results:
            avg = np.sum([results[target]['blending'][x] for x in results[target]['blending']], axis=0)
            avg /= len(results[target]['blending'])
            results.setdefault(target, {}).setdefault('avg', avg.tolist())

    # Rounding
    for target in results:
        for method in results[target]:
            if method == 'avg':
                results[target][method] = np.round(results[target][method], args.decimals).tolist()
            else:
                for fold in results[target][method]:
                    results[target][method][fold] = np.round(results[target][method][fold], args.decimals).tolist()

    # Output
    with open(args.out, 'w') as fh:
        json.dump(results, fh)


if __name__ == '__main__':
    main()



"""
Module containing categorical input features as well as output targets.
"""
from sklearn.base import TransformerMixin
import numpy as np

from allestm.features.base import NoFitMixin, LengthMixin


class Categorical:
    """
    Using the softmax activation function and the categorical crossentropy loss.
    """
    keras_activation = 'softmax'
    keras_loss = 'categorical_crossentropy'
    lightgbm_objective = 'multiclass'


class PropertiesCat(TransformerMixin, NoFitMixin, LengthMixin, Categorical):
    """
    Categorical properties:

    Class, Polarity, Charge, Asparagine or aspartic acid, Glutamine or glutamic acid,
    Leucine or Isoleucine, Hydrophobic, Aromatic, Aliphatic (non-aromatic), Small,
    Hydrophilic
    """
    _map = {
        'A': [0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0],
        'R': [1, 2, 2, 0, 0, 0, 0, 0, 0, 0, 1],
        'N': [2, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1],
        'D': [3, 3, 0, 1, 0, 0, 0, 0, 0, 0, 1],
        'C': [4, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
        'E': [3, 3, 0, 0, 1, 0, 0, 0, 0, 0, 1],
        'Q': [2, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1],
        'G': [0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0],
        'H': [5, 2, 2, 0, 0, 0, 0, 1, 0, 0, 1],
        'I': [0, 0, 1, 0, 0, 1, 1, 0, 1, 0, 0],
        'L': [0, 0, 1, 0, 0, 1, 1, 0, 1, 0, 0],
        'K': [1, 2, 2, 0, 0, 0, 0, 0, 0, 0, 1],
        'M': [4, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0],
        'F': [5, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0],
        'P': [6, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0],
        'S': [7, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1],
        'T': [7, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
        'W': [5, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0],
        'Y': [5, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0],
        'V': [0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0]
    }
    _feature_lengths = [8, 4, 3, 2, 2, 2, 2, 2, 2, 2, 2]

    length = sum(_feature_lengths)
    query = 'SELECT seqres FROM raw_data WHERE id=? ORDER BY resi'

    def __init__(self, onehot=True):
        """
        Categorical properties feature.

        Args:
            onehot: Transform the amino acid index to one-hot encoding.
        """
        self.onehot = onehot

    def transform(self, x):
        """
        If onehot=True, transforms the [n, 1]-shaped NumPy array containing the sequence to a [n, 20]-shaped
        NumPy array representing the one-hot encoded categorical features.

        Otherwise returns a [n, num_features]-shaped NumPy array converting the amino acid
        letters to feature values.

        Args:
            x: Sequence as [n, 1]-shaped NumPy array.

        Returns:
            [n, 20]-shaped NumPy array or [n, num_features]-shaped NumPy array.
        """
        if self.onehot:
            result = np.zeros((len(x), sum(self._feature_lengths)))
            for i, aa in enumerate(x[:, 0]):
                start = 0
                for feature_index, feature_length in enumerate(self._feature_lengths):
                    try:
                        onehot_index = start + self._map[aa][feature_index]
                        result[i, onehot_index] = 1.
                    except KeyError:
                        pass
                    start += feature_length
        else:
            result = np.zeros((len(x), len(self._feature_lengths)))
            for i, aa in enumerate(x[:, 0]):
                try:
                    result[i, :] = self._map[aa]
                except KeyError:
                    pass

        return result


class Sequence(TransformerMixin, NoFitMixin, LengthMixin, Categorical):
    """
    Sequence one-hot encoded.
    """
    _aas = ('A', 'R', 'N', 'D', 'C', 'Q', 'E', 'G', 'H', 'I', 'L', 'K', 'M', 'F', 'P', 'S', 'T', 'W', 'Y', 'V')
    _aa2index = {aa: index for index, aa in enumerate(_aas)}
    _index2aa = {index: aa for index, aa in enumerate(_aas)}

    length = len(_aas)
    query = 'SELECT seqres FROM raw_data WHERE id=? ORDER BY resi'

    def __init__(self, onehot=True):
        """
        Sequence feature.

        Args:
            onehot: Transform the amino acid index to one-hot encoding.
        """
        self.onehot = onehot

    def transform(self, x):
        """
        If onehot=True, transforms the [n, 1]-shaped NumPy array containing the sequence to a [n, 20]-shaped
        NumPy array representing the one-hot encoded sequence.

        Otherwise returns a [n, 1]-shaped NumPy array converting the amino acid
        letters to indices.

        Args:
            x: Sequence as [n, 1]-shaped NumPy array.

        Returns:
            [n, 20]-shaped NumPy array or [n, 1]-shaped NumPy array.
        """
        if self.onehot:
            result = np.zeros((len(x), len(self._aas)))
            for i, aa in enumerate(x[:, 0]):
                try:
                    result[i, self._aa2index[aa]] = 1.
                except KeyError:
                    pass
        else:
            result = np.zeros((len(x), 1))
            for i, aa in enumerate(x[:, 0]):
                try:
                    result[i, 0] = self._aa2index[aa]
                except KeyError:
                    pass

        return result

    def inverse_transform(self, x):
        """
        If onehot=True, transforms the [n, 20]-shaped NumPy array containing the one-hot encoded
        sequence to a [n, 1]-shaped NumPy array representing sequence.

        Otherwise returns a [n, 1]-shaped NumPy array converting the amino acid
        indices to letters.

        Args:
            x: One-hot encoded sequence as [n, 20]-shaped NumPy array.

        Returns:
            [n, 1]-shaped NumPy array.
        """
        if self.onehot:
            return np.array([[self._index2aa[r]] for r in x.argmax(axis=-1)])
        else:
            result = np.array((len(x), 1), dtype=str)
            for i, aa in enumerate(x[:, 0]):
                try:
                    result[i, 0] = self._index2aa[aa]
                except KeyError:
                    pass
            return result


class SecStruc(TransformerMixin, NoFitMixin, LengthMixin, Categorical):
    """
    Secondary structure,
    i.e.
    - [1, 0, 0] for helix.
    - [0, 1, 0] for sheet.
    - [0, 0, 1] for coil.

    or

    - [0] for helix.
    - [1] for sheet.
    - [2] for coil.
    """
    length = 3
    query = 'SELECT sec FROM raw_data where id=? ORDER BY resi'

    _sec2index = {'H': 0,
                  'G': 0,
                  'I': 0,
                  'E': 1,
                  'B': 1}
    _sec2onehot = {'H': [1., 0., 0.],
                   'G': [1., 0., 0.],
                   'I': [1., 0., 0.],
                   'E': [0., 1., 0.],
                   'B': [0., 1., 0.]}

    def __init__(self, onehot=True):
        """
        Init the transformer.

        Args:
            onehot: If True, one-hot encode the secondary structure.
        """
        self.onehot = onehot

    def _to_categorical_func(self, value, onehot):
        if onehot:
            return self._sec2onehot.get(value, [0., 0., 1.])
        else:
            return [self._sec2index.get(value, [2])]

    def transform(self, x):
        """
        Transforms the NumPy array [n, 1] to the [n, 3] categories.
        Args:
            x: [n, 1]-sized NumPy array (secondary structure).

        Returns:
            [n, 3] or [n,1]-sized NumPy array depending on onehot setting.
        """
        return np.array([self._to_categorical_func(v[0], self.onehot) for v in x])


class Topology(TransformerMixin, NoFitMixin, LengthMixin, Categorical):
    """
    ZCoordinates converted to three class topology but using reasonable TM segments,
    i.e.
    - [1, 0, 0, 0] for inside.
    - [0, 1, 0, 0] for TM.
    - [0, 0, 1, 0] for outside.
    - [0, 0, 0, 1] for a reentrant region.

    or

    - [0] for inside.
    - [1] for TM.
    - [2] for outside.
    - [3] for a reentrant region.
    """
    length = 4
    query = 'SELECT topo FROM raw_data where id=? ORDER BY resi'

    _topo2index = {'I': 0, 'T': 1, 'O': 2, 'R': 3}
    _topo2onehot = {'I': [1., 0., 0., 0.],
                    'T': [0., 1., 0., 0.],
                    'O': [0., 0., 1., 0.],
                    'R': [0., 0., 0., 1.]}

    def __init__(self, onehot=True):
        """
        Init the transformer.

        Args:
            onehot: If True, one-hot encode the topology.
        """
        self.onehot = onehot

    def _to_categorical_func(self, value, onehot):
        if onehot:
            return self._topo2onehot[value]
        else:
            return [self._topo2index[value]]

    def transform(self, x):
        """
        Transforms the NumPy array [n, 1] to the [n, 4] categories.
        Args:
            x: [n, 1]-sized NumPy array (topology).

        Returns:
            [n, 4] or [n,1]-sized NumPy array depending on onehot setting.
        """
        return np.array([self._to_categorical_func(v[0], self.onehot) for v in x])


class ZCoordinates(TransformerMixin, NoFitMixin, LengthMixin, Categorical):
    """
    ZCoordinates converted to three class topology, i.e.
    - [1, 0, 0] for inside.
    - [0, 1, 0] for TM.
    - [0, 0, 1] for outside.

    or

    - [0] for inside.
    - [1] for TM.
    - [2] for outside.
    """
    length = 3
    query = 'SELECT ca_z, thickness FROM raw_data JOIN proteins USING (id) where id=? ORDER BY resi'

    def __init__(self, onehot=True):
        """
        Init the transformer.

        Args:
            onehot: If True, one-hot encode the topology.
        """
        self.onehot = onehot

    @staticmethod
    def _to_categorical_func(value, thickness, onehot):
        if value < -thickness:
            if onehot:
                return 1., 0., 0.
            else:
                return [0.]
        elif value > thickness:
            if onehot:
                return 0., 0., 1.
            else:
                return [2.]
        else:
            if onehot:
                return 0., 1., 0.
            else:
                return [1.]

    def transform(self, x):
        """
        Transforms the NumPy array [n, 2] to the [n, 3] categories.
        Args:
            x: [n, 2]-sized NumPy array (z-coord and thickness tuples).

        Returns:
            [n, 3] or [n,1]-sized NumPy array depending on onehot setting.
        """
        return np.array([self._to_categorical_func(*v, self.onehot) for v in x])




"""
Module containing base classes for input features as well as output targets.
"""


class NoFitMixin:
    """
    Mixin class for not fittable transformers
    """
    def fit(self, x=None, y=None):
        """
        Does nothing.

        Args:
            x: [n, 1]-shaped NumPy array.
            y: [n, 1]-shaped NumPy array.

        Returns:
            Itself.
        """
        return self


class LengthMixin:
    """
    Implements the __len__ method returning the value of the
    class or instance variable length
    """
    def __len__(self):
        return self.length



"""
Module containing binary input features as well as output targets.
"""
from sklearn.base import TransformerMixin
from sklearn.preprocessing import StandardScaler, Binarizer

from allestm.features.base import LengthMixin, NoFitMixin


class Binary:
    """
    Using the sigmoid activation function and binary crossentropy loss.
    """
    keras_activation = 'sigmoid'
    keras_loss = 'binary_crossentropy'
    lightgbm_objective = 'binary'


class Bfactors(TransformerMixin, NoFitMixin, LengthMixin, Binary):
    """
    Scale bfactors according to z-score normalization and convert to binary according
    to threshold.
    """
    length = 1
    query = 'SELECT bfactor FROM raw_data WHERE id=? ORDER BY resi'

    def __init__(self, threshold=0.03, with_mean=True, with_std=True, copy=True):
        """
        Constructor.

        Args:
            threshold: Threshold used for binarization.
            with_mean: Substract mean for standard scaling.
            with_std: Divide by standard deviation for standard scaling.
            copy: Copy input.
        """
        self.threshold = threshold
        self.standard_scaler = StandardScaler(with_mean=with_mean, with_std=with_std, copy=copy)
        self.binarizer = Binarizer(threshold=threshold, copy=copy)

    def transform(self, x):
        """
        Transform the bfactor values by standard scaling and binarization.

        Args:
            x: [n, 1]-shaped NumPy array.

        Returns:
            [n, 1]-shaped NumPy array.
        """
        return self.binarizer.transform(self.standard_scaler.fit_transform(x))



import importlib


def name_to_feature(name):
    """
    Converts a feature name, e.g. continuous.Pssm to a class instance.
    Basically the name is appended to mllib.features to form
    mllib.features.continuous.Pssm, the module is loaded, and an instance created.
    Args:
        name: Name of the module and the class.

    Returns:
        Instance of the loaded class.
    """
    module_name, class_name = name.rsplit(".", maxsplit=1)

    module = importlib.import_module(module_name)
    cls = getattr(module, class_name)
    return cls()



from setuptools import setup, find_packages

VERSION = '1.1.2'

with open('README.md', 'r') as fh:
    long_description = fh.read()


setup(name='allestm',
      version=VERSION,
      description='Predicting various structural features of transmembrane proteins.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      pyhon_requires=">=3.6.0",
      url='http://github.com/phngs/allestm',
      author='Dr. Peter Hönigschmid',
      author_email='hoenigschmid.peter@gmail.com',
      license='MIT',
      packages=find_packages(),
      install_requires=[
            'pandas',
            'numpy',
            'keras',
            'tensorflow',
            'scikit-learn',
            'xgboost',
            'requests'
      ],
      entry_points={
            'console_scripts': ['allestm=allestm.allestm:main']
      },
      zip_safe=False,
      include_package_data=True
      )



"""Parse fasta and a3m files."""
import numpy as np
import pathlib
import logging


def parse_a3m(filename):
    """Parse a a3m fasta file.

    Parameters
    ----------
    filename : string
        The a3m file.

    Returns
    -------
    tuple :
        (Sequence in an [n, 1] np.array, MSA in an [1, 1] np.array)
    """
    content = pathlib.Path(filename).read_text()

    seq = ''
    for line in content.split('\n'):
        if line.startswith('>'):
            if len(seq) > 0:
                break
        else:
            seq += line.strip().upper().replace('-', '')

    return (np.array([[x] for x in list(seq)]), np.array([[content]]))



"""
Module containing continuous input features as well as output targets.
"""
import math
import re
from collections import Counter
from collections import defaultdict

import itertools
from sklearn.base import TransformerMixin

import numpy as np
from sklearn.preprocessing import StandardScaler

from allestm.features.base import NoFitMixin, LengthMixin


class Continuous:
    """
    Using the linear activation function and MAE loss.
    """
    keras_activation = 'linear'
    keras_loss = 'mean_absolute_error'
    lightgbm_objective = 'regression'


class RawValues(TransformerMixin, NoFitMixin, LengthMixin, Continuous):
    """
    Base class for directly using the database values without a specified query.
    """
    length = 1
    query = None

    @staticmethod
    def transform(x):
        """
        Just returns the original values from the database.
        Args:
            x: [n, 1]-shaped NumPy array.

        Returns:
            [n, 1]-shaped NumPy array.

        """
        return x


class RsaComplex(RawValues):
    """
    The relative solvent accessibility of a chain in a complex.
    """
    query = 'SELECT acc_complex FROM raw_data WHERE id=? ORDER BY resi'
    #keras_activation = 'sigmoid'


class RsaChain(RawValues):
    """
    The relative solvent accessibility of a chain without the remaining complex.
    """
    query = 'SELECT acc_chain FROM raw_data WHERE id=? ORDER BY resi'
    #keras_activation = 'sigmoid'


class RsaDiff(TransformerMixin, NoFitMixin, LengthMixin, Continuous):
    """
    The difference in relative solvent accessibility of the residues with and
    without the remaining complex.
    """
    length = 1
    query = 'SELECT acc_complex, acc_chain FROM raw_data WHERE id=? ORDER BY resi'
    #keras_activation = 'sigmoid'

    @staticmethod
    def transform(x):
        """
        Calcualte the absolute difference in RSA for residues with and without the
        remaining complex.
        Args:
            x: [n, 2]-shaped NumPy array.

        Returns:
            [n, 1]-shaped NumPy array.

        """
        return np.abs(x[:, 0] - x[:, 1])[:, None]


class PropertiesCont(TransformerMixin, NoFitMixin, LengthMixin, Continuous):
    """
    Continuous properties:

    Hydropathy, weight.
    """
    _map = {
        'A': [1.8, 89.094],
        'R': [-4.5, 174.203],
        'N': [-3.5, 132.119],
        'D': [-3.5, 133.104],
        'C': [2.5, 121.154],
        'E': [-3.5, 147.131],
        'Q': [-3.5, 146.146],
        'G': [-0.4, 75.067],
        'H': [-3.2, 155.156],
        'I': [4.5, 131.175],
        'L': [3.8, 131.175],
        'K': [-3.9, 146.189],
        'M': [1.9, 149.208],
        'F': [2.8, 165.192],
        'P': [-1.6, 115.132],
        'S': [-0.8, 105.093],
        'T': [-0.7, 119.119],
        'W': [-0.9, 204.228],
        'Y': [-1.3, 181.191],
        'V': [4.2, 117.148]
    }

    length = 2
    query = 'SELECT seqres FROM raw_data WHERE id=? ORDER BY resi'

    def __init__(self):
        """
        Continuous properties feature.
        """

    def transform(self, x):
        """
        Returns a [n, num_features]-shaped NumPy array converting the amino acid
        letters to feature values.

        Args:
            x: Sequence as [n, 1]-shaped NumPy array.

        Returns:
            [n, num_features]-shaped NumPy array.
        """
        result = np.zeros((len(x), self.length))
        for i, aa in enumerate(x[:, 0]):
            try:
                result[i, :] = self._map[aa]
            except KeyError:
                pass

        return result


class Pssm(TransformerMixin, NoFitMixin, LengthMixin, Continuous):
    """
    PSSM matrix calculated from a MSA.
    """
    _aas = ('A', 'R', 'N', 'D', 'C', 'Q', 'E', 'G', 'H', 'I', 'L',
            'K', 'M', 'F', 'P', 'S', 'T', 'W', 'Y', 'V')  # , '-')
    _aa2index = {aa: index for index, aa in enumerate(_aas)}

    length = len(_aas)
    query = 'SELECT msa FROM alignments where id=?'

    @classmethod
    def transform(cls, x):
        """
        Transforms the [1, 1]-shaped NumPy array to a [n, 20]-shaped one representing
        the PSSM.

        Args:
            x: [1,1]-shaped NumPy array containing the alignment in a3m format as a binary str.
            If the array is longer in any dimension, these are ignored.

        Returns:
            [n, 20]-shaped NumPy array representing the PSSM.
        """
        # Remove lowercase letters (gaps in query), remove headers
        msa = [re.sub('[a-z]', '', row)
               for row in x[0, 0].strip().split('\n') if not row.startswith('>')]

        # Transpose alignment (rows to cols)
        msa = list(map(list, zip(*msa)))

        pssm = np.zeros((len(msa), len(cls._aas)))

        # Count
        background_counter = Counter()
        for col_index, col in enumerate(msa):
            col_counter = Counter({aa: 1 for aa in cls._aas})
            col_counter.update(col)
            background_counter.update(col_counter)

            col_total_aas = sum(
                [v for k, v in col_counter.items() if k in cls._aa2index])
            for aa, aa_index in cls._aa2index.items():
                pssm[col_index, aa_index] = col_counter[aa] / col_total_aas

        # Calculate log ratios and normalize using tanh.
        background_total_aas = sum(
            [v for k, v in background_counter.items() if k in cls._aa2index])
        for col_index in range(len(msa)):
            for aa, aa_index in cls._aa2index.items():
                pssm[col_index, aa_index] /= background_counter[aa] / \
                    background_total_aas
                pssm[col_index, aa_index] = cls._normalize(
                    pssm[col_index, aa_index])

        return pssm

    @staticmethod
    def _normalize(x):
        return 1 / (1 + math.exp(-1 * math.log2(x)))


class ZCoordinates(TransformerMixin, NoFitMixin, LengthMixin, Continuous):
    """
    Normalized z-coordinates to a range of -1 to 1.
    """
    length = 1
    query = 'SELECT ca_z FROM raw_data where id=? ORDER BY resi'
    #keras_activation = 'tanh'

    def __init__(self, lower_bound=-25, upper_bound=25):
        """
        Constructor.

        Args:
            lower_bound: Lowest value for the z-coordinate.
            upper_bound: Highest value for the z-coordiante.
        """
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def _transform_func(self, v):
        v = max(self.lower_bound, min(self.upper_bound, v))
        return -1 + ((v - self.lower_bound) * 2 / (self.upper_bound - self.lower_bound))

    def transform(self, x):
        """
        Transforms the z-coordinate to a value between -1 and 1.

        Args:
            x: [n, 1]-shaped NumPy array.

        Returns:
            [n, 1]-shaped NumPy array.
        """

        return np.vectorize(self._transform_func)(x)

    def _inverse_transform_func(self, v):
        v = self.lower_bound + \
            (((v + 1) * (self.upper_bound - self.lower_bound)) / 2)
        return max(self.lower_bound, min(self.upper_bound, v))

    def inverse_transform(self, x):
        """
        Transforms transformed values back to original values (up to the lower and upper
        bounds).

        Args:
            x: [n, 1]-shaped NumPy array.

        Returns:
            [n, 1]-shaped NumPy array.
        """

        return np.vectorize(self._inverse_transform_func)(x)


class Bfactors(TransformerMixin, LengthMixin, NoFitMixin, Continuous):
    """
    Scale bfactors according to z-score normalization
    """
    length = 1
    query = 'SELECT bfactor FROM raw_data WHERE id=? ORDER BY resi'

    def __init__(self, with_mean=True, with_std=True, copy=True):
        self.standard_scaler = StandardScaler(
            with_mean=with_mean, with_std=with_std, copy=copy)

    def transform(self, x):
        """
        Transform the bfactor values by standard scaling.

        Args:
            x: [n, 1]-shaped NumPy array.

        Returns:
            [n, 1]-shaped NumPy array.
        """
        return self.standard_scaler.fit_transform(x)


class Angles(TransformerMixin, NoFitMixin, LengthMixin, Continuous):
    """
    Base class for angle transformation without a specified query.
    """
    length = 1
    query = None
    #keras_activation = 'tanh'

    @staticmethod
    def transform(x):
        """
        Transforms values from -pi to pi into a range of -1 to 1.
        Args:
            x: [n, 1]-shaped NumPy array.

        Returns:
            [n, 1]-shaped NumPy array.

        """
        return np.vectorize(lambda v: max(min(v / math.pi, 1.0), -1.0))(x)

    @staticmethod
    def inverse_transform(x):
        """
        Transforms values from -1 to 1 into a range of -180 to 180 degrees.
        Args:
            x: [n, 1]-shaped NumPy array.

        Returns:
            [n, 1]-shaped NumPy array.

        """
        return np.vectorize(lambda v: max(min(v, 1.0), -1.0) * 180.0)(x)
        # return np.vectorize(lambda v: max(min(v, 1.0), -1.0) * math.pi)(x)


class PhiAngles(Angles, LengthMixin):
    """
    Implementation of the Angles feature including a query for phi angles.
    """
    query = 'SELECT phi FROM raw_data WHERE id=? ORDER BY resi'


class PsiAngles(Angles, LengthMixin):
    """
    Implementation of the Angles feature including a query for psi angles.
    """
    query = 'SELECT psi FROM raw_data WHERE id=? ORDER BY resi'


class Thickness(TransformerMixin, NoFitMixin, LengthMixin, Continuous):
    """
    Returns half thickness of the membrane.
    """
    length = 1
    query = 'SELECT thickness FROM raw_data JOIN proteins USING (id) where id=? ORDER BY resi'

    @staticmethod
    def transform(x):
        """
        Just returns the thicknesses.

        Args:
            x: [n, 1]-shaped NumPy array.

        Returns:
            [n, 1]-shaped NumPy array.

        """
        return x
